/*
 *  ofxQTKitMoviePlayer example
 *
 * Created by James George, http://www.jamesgeorge.org
 * over a long period of time for a few different projects in collaboration with
 * FlightPhase http://www.flightphase.com 
 * and the rockwell group lab http://lab.rockwellgroup.com
 *
 **********************************************************
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * ----------------------
 * 
 * ofxQTKitVideoPlayer provides QTKit/CoreVideo accelerated movie playback
 * for openFrameworks on OS X
 * 
 * This class replaces almost all of the functionality of ofVideoPlayer on OS X
 * and uses the modern QTKit and CoreVideo libraries for playback
 *
 * Special Thanks to Marek Bereza for his initial QTKitVideoTexture
 * http://www.mrkbrz.com/
 *
 * Thanks to Anton Marini for help wrangling buffer contexts
 * http://vade.info/ 
 * some additions/modifications by Brian Chasalow - this is an extremely experimental/application-specific fork! brian@chasalow.com
 */

#include "ofxQTKitVideoPlayer.h"
#include "MonoCallbacks.h"

ofxQTKitVideoPlayer::ofxQTKitVideoPlayer()
{
	moviePlayer = NULL;
	moviePixels = NULL;
	bNewFrame = false;
	duration = 0;
	nFrames = 0;
    storedMovieFileName = "";
    loopState = OF_LOOP_NONE;
    videoHasEnded = false;
    speed = 0;
    volume = 0;
    iAmLoading = false;
    iAmLoaded = false;
    firstLoad = true;
}


ofxQTKitVideoPlayer::~ofxQTKitVideoPlayer()
{
	close();	
}

void ofxQTKitVideoPlayer::pseudoUpdate(){
    
    
}

void ofxQTKitVideoPlayer::pseudoDraw(int myint, int myotherint){
    
    
}

bool ofxQTKitVideoPlayer::loadMovie(string movieFilePath, int mode)
{
    NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];

	
    if(movieFilePath == ""){
                NSLog(@"Load BLOCKED by null name");
		[pool drain];
        return false;
    }
    
	if(mode < 0 || mode > 2){
		ofLog(OF_LOG_ERROR, "ofxQTKitVideoPlayer -- Error, invalid mode specified for");
		[pool drain];
		return false;
	}
   
    
    if(moviePlayer != NULL){
		close();
	}
    if(!iAmLoading){

	if(pool == NULL)
	pool = [[NSAutoreleasePool alloc] init];
    

	
	bool useTexture = (mode == OFXQTVIDEOPLAYER_MODE_TEXTURE_ONLY || mode == OFXQTVIDEOPLAYER_MODE_PIXELS_AND_TEXTURE);
	bool usePixels  = (mode == OFXQTVIDEOPLAYER_MODE_PIXELS_ONLY  || mode == OFXQTVIDEOPLAYER_MODE_PIXELS_AND_TEXTURE);
//	         NSLog(@"inside: mode: %i, useTexture, usePixels. %@, %@",mode, (useTexture ? @"YES" : @"NO"),( usePixels? @"YES" : @"NO"));
	moviePlayer = [[QTKitMovieRenderer alloc] init];
	
	//movieFilePath = ofToDataPath(movieFilePath, false);
	BOOL success = [moviePlayer loadMovie:[NSString stringWithCString:ofToDataPath(movieFilePath, false).c_str() encoding:NSUTF8StringEncoding] 
							 allowTexture:useTexture 
							  allowPixels:usePixels];

    
    storedMovieFileName = movieFilePath;

    

    
	if(!success) {
		ofLog(OF_LOG_ERROR, "ofxQTKitVideoPlayer -- Loading file " + movieFilePath + " failed");
		[moviePlayer release];
		moviePlayer = NULL;

	}
    else{
        

        iAmLoading = true;
        iAmLoaded = false; 
    }
            
            
	[pool drain];
	
	return success;
        
        
        }
	
	[pool drain];
    return false;

}

void ofxQTKitVideoPlayer::closeMovie()
{
	close();
}

bool ofxQTKitVideoPlayer::isLoaded()
{
	return moviePlayer != NULL;
}


string ofxQTKitVideoPlayer::getCurrentlyPlaying() {
    return storedMovieFileName;
}



void ofxQTKitVideoPlayer::close()
{
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
    storedMovieFileName = "";	
	if(moviePlayer != NULL){
		[moviePlayer release];
		moviePlayer = NULL;
	}
	
	if(moviePixels != NULL){
		delete moviePixels;
		moviePixels = NULL;
	}
	
	duration = 0;
	nFrames = 0;

	[pool drain];	
}

void ofxQTKitVideoPlayer::pause()
{
    speed = getSpeed();
	setSpeed(0);
}

//bool ofxQTKitVideoPlayer::isPaused()
//{
//	return getSpeed() == 0.0;
//}

void ofxQTKitVideoPlayer::setPaused(bool myPause){
    if(myPause){
        isPaused = true;

        if(iAmLoaded)
		pause();
    }
    else{
        isPaused = false;

        
        if(iAmLoaded){
        
        if(speed != 0)
            setSpeed(speed);
        else 
            setSpeed(1);
            
        }
    }
}

void ofxQTKitVideoPlayer::setSpeed(float rate)
{
	if(moviePlayer == NULL) return;
	
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	
    speed = rate;
	[moviePlayer setRate:rate];

	[pool drain];	
}

void ofxQTKitVideoPlayer::play()
{	
	if(moviePlayer == NULL) return;

	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	
	[moviePlayer setRate: 1.0];
	
	[pool drain];
}

void ofxQTKitVideoPlayer::idleMovie()
{
	update();
}

bool ofxQTKitVideoPlayer::update()
{
    if(iAmLoading){
		if(moviePlayer.loadState == -1000){
			handleVideoLoaded(NULL);
			moviePlayer.loadState = -999;
		}
			
        if(moviePlayer.loadState >=2000){
            iAmLoading = false;
            iAmLoaded = true;
            videoHasEnded = false;
            speed = 1;
            setLoopState(OF_LOOP_NONE);
            setPaused(isPaused);
            setVolume(volume);
            duration = moviePlayer.duration;
            nFrames = moviePlayer.frameCount;
            width = moviePlayer.movieSize.width;
            height = moviePlayer.movieSize.height;
            handleVideoLoaded(this);
            firstLoad = false;
        }
        return false;
	}

    
    
	if(moviePlayer == NULL) return false;

	bNewFrame = [moviePlayer update];
	if (bNewFrame) {
		bHavePixelsChanged = true;
	}
    resetToZeroIfDone();
    return bNewFrame;
}

void ofxQTKitVideoPlayer::resetToZeroIfDone(){
	if(getIsMovieDone()){
     //   NSLog(@"MOVIE ENDED");
		if(!isPaused && !iAmLoading){
       // NSLog(@"MOVIE NOT PAUSED, NOT LOADING");            
            if(!videoHasEnded){
         //       NSLog(@"MOVIE NOT ENDED");
                //you only get one shot to tell us that the movie has ended, for each time the movie successfully loads.
                videoHasEnded = true;
           //     NSLog(@"MOVIE ENDED");

                
                //provide callback if the video ends
                handleVideoEnded(this);
                //reset to zero if you're set to loop normal mode.
                if(loopState == OF_LOOP_NORMAL){
                    setPosition(0.0);
                    setPaused(isPaused);
                    
                }
            
            }
		}
        
	}	
}


int ofxQTKitVideoPlayer::getMovieLoopState()
{
	return loopState;
}


//bool ofxQTKitVideoPlayer::getMovieLoopState()
//{
//	if(moviePlayer == NULL) return NO;
//	
//	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
//	
//	bool loops = moviePlayer.loops;
//	
//	[pool drain];
//	
//	return loops;
//}


bool ofxQTKitVideoPlayer::isFrameNew()
{
	return bNewFrame;
}
		
void ofxQTKitVideoPlayer::bind()
{
	if(moviePlayer == NULL || !moviePlayer.useTexture) return;
	
	[moviePlayer bindTexture];	
}

void ofxQTKitVideoPlayer::unbind()
{
	if(moviePlayer == NULL || !moviePlayer.useTexture) return;
	
	[moviePlayer unbindTexture];
}

void ofxQTKitVideoPlayer::draw(float x, float y)
{
	if(moviePlayer == NULL || !moviePlayer.useTexture) return;
	
	draw(x,y, moviePlayer.movieSize.width, moviePlayer.movieSize.height);
}

void ofxQTKitVideoPlayer::draw(float x, float y, float w, float h)
{
	if(moviePlayer == NULL || !moviePlayer.useTexture) return;
	
	[moviePlayer draw:NSMakeRect(x, y, w, h)];
}

unsigned char* ofxQTKitVideoPlayer::getPixels()
{
	if(moviePlayer == NULL || !moviePlayer.usePixels) {
		return NULL;
	}
		
	if(moviePixels == NULL){
		moviePixels = new unsigned char[int(moviePlayer.movieSize.width) * int(moviePlayer.movieSize.height) * 4];
	}
		
	//don't get the pixels every frame if it hasn't updated
	if(bHavePixelsChanged){
		[moviePlayer pixels:moviePixels];
		bHavePixelsChanged = false;
	}
	
	return moviePixels;
}

void ofxQTKitVideoPlayer::setPosition(float pct)
{
	if(moviePlayer == NULL) return;

	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];

    if(pct == 0.0){
        videoHasEnded = false;
    }
	moviePlayer.position = pct;
	
	[pool drain];
}

void ofxQTKitVideoPlayer::setVolume(int myVolume)
{
	if(moviePlayer == NULL) return;
		
    volume = myVolume;
    
    if(iAmLoaded){
	moviePlayer.volume = volume;
    }

}

void ofxQTKitVideoPlayer::setFrame(int frame)
{
	if(moviePlayer == NULL) return;

	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	
	moviePlayer.frame = frame % moviePlayer.frameCount;
	
	[pool drain];
	
}

int ofxQTKitVideoPlayer::getCurrentFrame()
{
	if(moviePlayer == NULL) return 0;

	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	
	int currentFrame = moviePlayer.frame;
	
	[pool drain];	
	
	return currentFrame;	
}

int ofxQTKitVideoPlayer::getTotalNumFrames()
{
	return nFrames;
}

void ofxQTKitVideoPlayer::setLoopState(bool loops)
{
	if(moviePlayer == NULL) return;
	
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	
	moviePlayer.loops = loops;
	
	[pool drain];
}

void ofxQTKitVideoPlayer::setLoopState(int ofLoopState)
{
	if(ofLoopState == OF_LOOP_NONE){
		setLoopState(false);
	}
	else if(ofLoopState == OF_LOOP_NORMAL){
		setLoopState(true);
	}
	
	//TODO support OF_LOOP_PALINDROME
}



float ofxQTKitVideoPlayer::getSpeed()
{
//	if(moviePlayer == NULL) return 0;
//	
//	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
//	
//	float rate = moviePlayer.rate;
//	
//	[pool drain];
//	
	return speed;
}

float ofxQTKitVideoPlayer::getDuration()
{
	return duration;
}

float ofxQTKitVideoPlayer::getPositionInSeconds()
{
	return getPosition() * duration;
}

float ofxQTKitVideoPlayer::getPosition()
{
	if(moviePlayer == NULL) return 0;
	
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
	
	float pos = moviePlayer.position;
	
	[pool drain];
	
	return pos;
}

bool ofxQTKitVideoPlayer::getIsMovieDone()
{
	if(moviePlayer == NULL) return false;
	
	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];	

	bool isDone = moviePlayer.isFinished;
	
	[pool drain];
	
	return isDone;
}

int ofxQTKitVideoPlayer::getWidth()
{
	return moviePlayer.movieSize.width;
}

int ofxQTKitVideoPlayer::getHeight()
{
	return moviePlayer.movieSize.height;
}

